/**
*	Author:	Elizabeth Labelle
*	ID:			17417
*	Project:	Boids Simulation
*	Date:		December 2010
*
*	Description:	A Boid is a single boid within a flock. It is spawned by a BoidManager, which
*					keeps track of variables influencing flock and individual Boid behaviors.
*					Each Boid moves in the world based on the three rules of flocking: Cohesion, Seperation, 
*					and Alignment. Each Boid also tries to stay within world bounds.
*					Boids can be hunted and eaten by AttackBoids.
*/
using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public class Boid : MonoBehaviour 
{

	/**Set by the BoidManager when it creates each Boid. 
	*	References the BoidManager with flock statistics. 
	**/
	internal BoidManager controller;
	
	/**
	*	Called when the Boid object is first instantiated by the Unity Engine.
	*	Calls the Coroutine which calculates new movement vectors every 200-600 ms
	*/
	void Start()
	{
		if(tag == "" || tag == null)
			tag = "Boid";
		//using a coroutine allows the routine to continue running in a loop, but start to return
		StartCoroutine(steering());
	}
	
	/**
	*	steering() is a Unity coroutine.
	*	Method calculates movement vectors based on the three flocking rules:
	*		1) Cohesion - Each boid tries to move to the center of the flock
	*		2) Seperation - Each boid tries to stay a set distance away from other Boids
	*		3) Alignment - Each Boid tries to steer towards the average heading of other Boids
	*	Boids also attempt to stay in bounds of the world, and especially try to avoid the ground (to
	*	avoid intersecting with the ground geometry.)
	*/
	IEnumerator steering()
	{
		if(controller == null)
		{ //if controller is null, the Boid doesn't have a BoidManager
			//so don't simulate behavior for it
			return false;
		}
		
		Vector3 rule1, rule2, rule3, rule4, rule5;
		
		//loop while the program is running (always)
		while(true)
		{
			//Calculate vectors for the three rules + bounds checking
			//and add additional weights to the results
			rule1 = stayTogether() * 10;
			rule2 = dontCollide();
			rule3 = matchVelocity() * 50;
			rule4 = tendStayInBounds() * 10;
			rule5 = tendAvoidGround() * 100;//adds more weight to specifically avoiding the ground
			
			rigidbody.velocity = rigidbody.velocity + rule1 + rule2 + rule3 + rule4 + rule5;
			//add some "noise" so all boids don't act the same way
			rigidbody.velocity += new Vector3(Random.Range(-0.2f, 0.2f),
															Random.Range(-0.2f, 0.2f), 
															Random.Range(-0.2f, 0.2f));
			
			
			//keep velocity within bounds set by controller
			float speed = rigidbody.velocity.magnitude;
			
			if(speed > controller.maxVelocity)
				rigidbody.velocity = rigidbody.velocity.normalized * controller.maxVelocity;
			else if(speed < controller.minVelocity)
				rigidbody.velocity = rigidbody.velocity.normalized * controller.minVelocity;
			
			//turn to face the new velocity
			adjustFacing(rigidbody.velocity);
			
			//wait 200-600 ms to recalculate movement vectors
			float waitTime = Random.Range(0.2f, 0.6f);
			yield return new WaitForSeconds(waitTime);
		}
	}
	
	/**
	*	Returns a vector that will lead the boid towards the center of the flock.
	*	Vector returned is 1% of the actual vector needed to reach the center.
	*/
	Vector3 stayTogether()
	{
		Vector3 vec = Vector3.zero;
		
		vec = controller.boidCenter;
		
		return (vec - rigidbody.position) / 100;//mov boid 1% closer to center of flock
	}
	
	/**
	*	Iterates through every boid in the BoidManager's boidList. Returns the vector that will keep
	*	the boid a set distance from all other boids in the flock.
	*	Uses BoidManager's minDist value to determine how far away one boid must be from another.
	*/
	Vector3 dontCollide()
	{
		Vector3 vec = Vector3.zero;
		
		foreach(Boid b in controller.boidList)
		{
			if(b != this)
			{
				if(Mathf.Abs(b.rigidbody.position.x - this.rigidbody.position.x) < controller.minDist)
					vec.x = vec.x - (b.rigidbody.position.x - this.rigidbody.position.x);
				
				if(Mathf.Abs(b.rigidbody.position.y - this.rigidbody.position.y) < controller.minDist)
					vec.y = vec.y - (b.rigidbody.position.y - this.rigidbody.position.y);
				
				if(Mathf.Abs(b.rigidbody.position.z - this.rigidbody.position.z) < controller.minDist)
					vec.z = vec.z - (b.rigidbody.position.z - this.rigidbody.position.z);
			}
		}
		
		return vec;
	}
	
	/**
	*	Method examines every boid in the BoidManager's boidList and returns a fraction of the average velocity.
	*/
	Vector3 matchVelocity()
	{
		Vector3 vec = Vector3.zero;
		
		vec = controller.boidVelocity;
		
		return (vec - rigidbody.velocity) / 100;//return 1% of the average velocity
	}
	
	/**
	*	Encourages boids to stay in bounds by returning a vector in the opposite direction of a boundary
	*	when the boid gets too close.
	*	Boids can still travel outside the boundary if their velocity is higher than the returned offset, so
	*	the boundary used for checking is smaller than the world boundary. This looks more natural than
	*	boids bouncing off of invisible walls.
	*/
	Vector3 tendStayInBounds()
	{
		GameObject worldObject = GameObject.FindWithTag("WorldData");
		World worldData = worldObject.GetComponent<World>();
		
		Vector3 vec = Vector3.zero;
		
		if(rigidbody.position.x < worldData.xMin)
			vec.x = 1;
		else if(rigidbody.position.x > worldData.xMax)
			vec.x = -1;
		
		if(rigidbody.position.y < worldData.yMin)
			vec.y = 1;
		else if(rigidbody.position.y > worldData.yMax)
			vec.y = -1;
		
		if(rigidbody.position.z < worldData.zMin)
			vec.z = 1;
		else if(rigidbody.position.z > worldData.zMax)
			vec.z = -1;
		
		return vec;
	}
	
	/**
	*	Encourages boids to avoid the ground by returning a vector in the opposite direction when they get too close.
	*	Uses the BoidManager's minDistGround variable to decide how close the boids can get before being repelled.
	*/
	Vector3 tendAvoidGround()
	{
		Vector3 avoidVector = Vector3.zero;
		
		GameObject ground = GameObject.FindWithTag("Ground");
		Vector3 avoidLocation = ground.transform.position;
		
		if(rigidbody.position.y < controller.minDistGround) 
		{
			avoidVector.y = avoidLocation.y - rigidbody.position.y / 10;
			avoidVector = avoidVector * -1;//go in the opposite direction
		}
		
		return avoidVector;
	}
	
	/**
	*	Turns a boid to point in the direction of its current velocity.
	*/
	void adjustFacing(Vector3 destination)
	{
		if(destination != Vector3.zero)
			transform.rotation = Quaternion.LookRotation(destination);
	}
	
	/**
	*	Removes the Boid from the BoidManager and deletes the Boid object
	*/
	public void beEaten()
	{
		controller.removeBoid(this);
		Destroy(this);//destroys object, not GameObject in scene!
	}
}
